#include <stdio.h>
#include "itemlist.h"
#include "local.h"
#include "gramactions.h"

char* mysprintf (char *format, ...);
int  new_state (void);

int run_exists = 0;

/****************************************************/
void InitAll (void)
{
   fprintf (stdout, "#include \"loft_c.h\"\n");
}

/*************** cooperate ***************/
char *Cooperate (void)
{
   return mysprintf ("LOFT_STATE(%d) {LOFT_GOTO_NEXT;}",new_state ());
}

/*************** halt ***************/
char *Halt (void)
{
   int st = new_state ();
   return mysprintf ("LOFT_STATE(%d) {LOFT_GOTO(%d);}",st,st);
}

/*************** return ***************/
char *Return (void)
{
   return mysprintf ("LOFT_STATE(%d) {LOFT_IMMEDIATE(-1);}",new_state ());
}
   
/*************** C Statements ***************/
char *Cstat(char *s)
{
   return mysprintf ("LOFT_STATE(%d) {%s}",new_state (),s);
}

/*************** await ***************/
static char *await_shape = "\
LOFT_STATE(%d) {local(%s)=%s;}\n\
LOFT_STATE(%d) {\n\
   event_t event = local(%s);\n\
   if (event->scheduler != current_scheduler ()) {\n\
      self ()->code = EBADLINK;\n\
   } else if (!is_present (event)) {\n\
      register_in_event (event);\n\
      return _WAIT;\n\
   }\n\
}";

static char *await_n_shape = "\
LOFT_STATE(%d) {\
   local(%s) = %s; self ()->deadline = current_scheduler()->instant + %s;\
}\n\
LOFT_STATE(%d) {\n\
   scheduler_t sched = current_scheduler ();\n\
   event_t event = local(%s);\n\
   if (event->scheduler != sched) {\n\
      self ()->code = EBADLINK;\n\
   } else if (self ()->deadline <= sched->instant) {\n\
      self ()->code = ETIMEOUT;\n\
   } else if (is_present (event)) {\n\
      self ()->code = OK;\n\
   } else {\n\
      register_in_event (event);\n\
      register_in_timer (sched);\n\
      return _WAIT;\n\
   }\n\
}";

char *Await (item_list_t list)
{
   char *sigexp, *numexp, *evt;
   item_cell_t cell;
   int len = _length_of_item_list (list);
   int s1 = new_state () ,s2 = new_state ();
   
   if (len != 1 && len != 2){
      return "bad_number_of_arguments_to_await\n";
   }
   cell = _first_cell (list);
   
   sigexp = _content_of_cell (cell);
   evt = next_local_var ();
   add_local_data ("event_t",evt);

   if (len == 1) return mysprintf(await_shape,s1,evt,sigexp,s2,evt); 

   numexp = _content_of_cell (_next_cell (cell));
   
   return mysprintf (await_n_shape,
		     s1,
                     evt,sigexp,numexp,
		     s2,
		     evt);
}

/*************** join ***************/
static char *join_shape = "\
LOFT_STATE(%d) {local(%s)=%s;}\n\
LOFT_STATE(%d) {\
   thread_t thread = local(%s);\n\
   event_t event = thread->term_signal;\n\
   if (event->scheduler != current_scheduler ()) {\n\
      self ()->code = EBADLINK;\n\
   } else if (thread->state != _TERM && !is_present (event)) {\n\
      register_in_event (event);\n\
      return _WAIT;\n\
   }\n\
}";

static char *join_n_shape = "\
LOFT_STATE(%d) {\
   local(%s)=%s; self ()->deadline = current_scheduler()->instant + %s;\
}\n\
LOFT_STATE(%d) {\n\
   scheduler_t sched = current_scheduler ();\n\
   thread_t thread = local(%s);\n\
   event_t event = thread->term_signal;\n\
   if (event->scheduler != sched) {\n\
      self ()->code = EBADLINK;\n\
   } else if (self ()->deadline <= sched->instant) {\n\
      self ()->code = ETIMEOUT;\n\
   } else if (thread->state == _TERM || is_present (event)) {\n\
      self ()->code = OK;\n\
   } else {\n\
      register_in_event (event);\n\
      register_in_timer (sched);\n\
      return _WAIT;\n\
   }\n\
}";


char *Join (item_list_t list)
{
   char *thexp, *numexp, *th;
   item_cell_t cell;
   int len = _length_of_item_list (list);
   int s1 = new_state () ,s2 = new_state ();
   
   if (len != 1 && len != 2){   
      return "bad_number_of_arguments_to_join";
   }
   cell = _first_cell (list);
   
   thexp = _content_of_cell (cell);
   th = next_local_var ();
   add_local_data ("thread_t",th);
   
   if (len == 1) {
      return mysprintf (join_shape,
			s1,th,thexp,
			s2,
			th);
   }
   numexp = _content_of_cell (_next_cell (cell));
   
   return mysprintf (join_n_shape,
		     s1,
		     th,thexp,numexp,
		     s2,
		     th);
}

/*************** get_value ***************/
static char *get_shape = "\
LOFT_STATE(%d) {\
   local(%s)=%s;\
   local(%s)=%s;\
   local(%s)=%s;\
}\n\
LOFT_STATE(%d) {\n\
   scheduler_t sched = current_scheduler ();\n\
   event_t event = local(%s);\n\
   int n = local(%s);\n\
   int len = event_values_length (event);\n\
   if (event->scheduler != sched) {\n\
      self ()->code = EBADLINK;\n\
   } else if (is_present (event) && len > n) {\n\
      (*local(%s)) = nth_of_event_values (event,n);\n\
      self ()->code = OK;\n\
   } else if (sched->eoi) {\n\
      self ()->code = ENEXT;\n\
      LOFT_GOTO_NEXT;\n\
   } else return _CONT;\n\
}";


char *Get_Value (item_list_t list)
{
   char *eventexp, *numexp, *resultexp, *event, *num, *result;
   int s1 = new_state () ,s2 = new_state ();
   item_cell_t cell;
   if (_length_of_item_list (list) != 3){
      return "bad_number_of_arguments_to_get_value";
   }
   
   cell = _first_cell (list);
   eventexp = _content_of_cell (cell);
   event = next_local_var ();
   add_local_data ("event_t",event);

   numexp = _content_of_cell (_next_cell (cell));
   num =  next_local_var ();
   add_local_data ("int",num);
   
   resultexp = _content_of_cell (_next_cell (_next_cell (cell)));
   result =  next_local_var ();
   add_local_data ("void**",result);

   return mysprintf (get_shape,
		     s1,
		     event,eventexp,
		     num,numexp,
		     result,resultexp,
		     s2,
		     event,
		     num,
		     result);
}

/*************** link, unlink ***************/
char *Unlink (void)
{
   return "unlink_not_possible();";
}

static char *link_shape = "\
LOFT_STATE(%d) {\
add_to_unlink (self ()->scheduler,self ());\
add_to_link (%s,self ()); \
LOFT_GOTO_NEXT;}";

char *Link (char *sched)
{
   return mysprintf (link_shape,new_state(),sched);
}

/*************** if ***************/
static char *if_shape = "\
LOFT_STATE(%d) {if (!(%s)) LOFT_IMMEDIATE (%d);}\n\
%s\nLOFT_STATE(%d) {LOFT_IMMEDIATE(%d);}\n\
%s\nLOFT_STATE(%d)";

char *If (int state,char *cond,char *th,int elstate,char *el)
{
   int end = new_state ();
   return mysprintf (if_shape,state,cond,elstate+1,th,elstate,end,el,end);
}

/*************** while ***************/
static char *while_shape = "\
LOFT_STATE(%d) {if (!(%s)) LOFT_IMMEDIATE (%d);}\n\
%s\n\
LOFT_STATE(%d) {LOFT_IMMEDIATE (%d);}\n\
LOFT_STATE(%d)";

char *While (int state,char *exp,char *body)
{
   int redo,end;
   redo = new_state ();
   end  = new_state();
   return mysprintf (while_shape,state,exp,end,body,redo,state,end);
}

/*************** repeat ***************/
int InitRepeat (void)
{
   int res = new_state ();
   new_state (); // reserve one state number
   return res;
}

static char *repeat_shape = "\
LOFT_STATE(%d) {local(%s) = %s;}\n\
LOFT_STATE(%d) {if (local(%s) <= 0) LOFT_IMMEDIATE(%d);}\n\
%s\n\
LOFT_STATE(%d) {local(%s)--; LOFT_IMMEDIATE (%d);}\n\
LOFT_STATE(%d)";

char *Repeat (int st,char *num,char *body)
{
   char *count = next_local_var ();
   int   decr  = new_state ();
   int   end   = new_state ();

   add_local_data ("int",count);
   
   return mysprintf (repeat_shape,
		     st,
		     count,num,
		     st+1,count,end,
		     body,
		     decr,count,st+1,
		     end);
}

/*************** run ***************/
static char *run_shape = "\
LOFT_STATE(%d) {%s}\n\
LOFT_STATE(%d) {\n\
   thread_t called = self ()->called;\n\
   event_t event = called->term_signal;\n\
   if (event->scheduler != current_scheduler ()) {\n\
      self ()->code = EBADLINK;\n\
   } else if (called->state != _TERM && !is_present (event)) {\n\
      register_in_event (event);\n\
      return _WAIT;\n\
   }\n\
   called = NULL;\n\
}";

static char *creation_shape = "self ()->called = %s_create (%s);";

char *Run (char *name,char *params)
{
   int s1 = new_state (),
       s2 = new_state ();
   run_exists = 1;
   return mysprintf (run_shape,
		     s1,mysprintf (creation_shape,name,params),
		     s2);
}

/*****************************************/
