#ifndef _THREAD_H
#define _THREAD_H

#define MAGIC 12345

typedef char (*automaton_t) ();
typedef void (*finalizer_t) ();

typedef struct _thread_t
{
   automaton_t           automaton;
   finalizer_t           finalizer;   
   void                 *locals;
   
   char                  state;
   
   char                  suspended;
   struct _event_t      *term_signal;
   struct _thread_t     *called;
   char                  code;
   
   struct _scheduler_t  *scheduler;   

   struct _event_t      *awaited_event;
   int                   deadline;
   item_cell_t           event_cell;
   item_cell_t           timer_cell;
   
   int                   magic;
}
*thread_t;

thread_t make_thread         (struct _scheduler_t*,
			      automaton_t,
			      finalizer_t,
			      void *locals);

thread_t self                (void);

int      reset_thread_state  (thread_t);
void     deallocate_thread   (thread_t);

int      fireable            (thread_t);
void     fire_thread         (thread_t);
void     finalize_thread     (thread_t);
void     normal_termination  (thread_t);

void     unregister_in_timer (thread_t);
void     unregister_in_event (thread_t);

#endif /* _THREAD_H */

