#include <stdio.h>
#include <string.h>
#include "itemlist.h"
#include "local.h"
#include "gramactions.h"
#include "loft_ft.h"

extern int native;
char* mysprintf(char *format, ...);
int  new_state (void);

int run_exists = 0;

/****************************************************/
void InitAll (void)
{
   fprintf (stdout, "#include \"fthread_macros.h\"\n");
}

/*************** cooperate ***************/
char *Cooperate(void)
{
   if (native) return mysprintf ("ft_thread_cooperate ();");
   return mysprintf ("STATE (%d) {GOTO_NEXT;}",new_state ());
}

/*************** halt ***************/
char *Halt(void)
{
   int st;
   if (native) return mysprintf ("while(1) ft_thread_cooperate ();");
   st = new_state ();
   return mysprintf ("STATE (%d) {GOTO(%d);}",st,st);
}

/*************** return ***************/
char *Return(void)
{
   if (native) return "return;";
   return mysprintf ("STATE (%d) {IMMEDIATE(-1);}",new_state ());
}
   
/*************** C Statements ***************/
char *Cstat(char *s)
{
   if (native) return mysprintf ("{%s}",s);
   return mysprintf ("STATE (%d) {%s}",new_state (),s);
}

/*************** await ***************/
static char *nat_await_shape = "\
_set_thread_return_code (_self,ft_thread_await (%s));";

static char *nat_await_n_shape = "\
_set_thread_return_code (_self,ft_thread_await_n (%s,%s));";

char *Await(item_list_t list)
{
   char *sig, *num;
   item_cell_t cell;
   int len = _length_of_item_list (list);
   
   if (len != 1 && len != 2){
      return "error: bas number of arguments to await\n";
   }
   cell = _first_cell (list);
   sig = _content_of_cell (cell);

   if (len == 1){
      if (native) return mysprintf  (nat_await_shape,sig);
      return mysprintf("STATE_AWAIT (%d,%s) {}",new_state (),sig);
   }
   num = _content_of_cell (_next_cell (cell));

   if (native) return mysprintf (nat_await_n_shape,sig,num);
   return mysprintf ("STATE_AWAIT_N (%d,%s,%s) {}",new_state (),sig,num);
}

/*************** join ***************/
static char *join_shape = "\
STATE (%d) {local(%s) = %s;}\n\
STATE_JOIN (%d,local(%s)) {}";

static char *join_n_shape = "\
STATE (%d) {local(%s) = %s;}\n\
STATE_JOIN_N (%d,local(%s),%s) {}";

static char *nat_join_shape = "\
_set_thread_return_code (_self,ft_thread_join (%s));";

static char *nat_join_n_shape = "\
_set_thread_return_code (_self,ft_thread_join_n (%s,%s));";

char *Join(item_list_t list)
{
   char *th, *num, *loc;
   item_cell_t cell;
   int len = _length_of_item_list (list);
   int comp, join;
   if (len != 1 && len != 2){
      return "error: bas number of arguments to join\n";
   }
   cell = _first_cell (list);
   th = _content_of_cell (cell);

   if (native) {
      if (len == 1) return mysprintf (nat_join_shape,th);
      num = _content_of_cell (_next_cell (cell));
      return mysprintf (nat_join_n_shape,th,num);
   }
   
   loc = next_local_var ();   
   add_local_data ("ft_thread_t",loc);

   comp = new_state ();
   join = new_state ();

   if (len == 1){
      return mysprintf (join_shape,comp,loc,th,join,loc);
   }
   num = _content_of_cell (_next_cell (cell));
   return mysprintf (join_n_shape,comp,loc,th,join,loc,num);
}

/*************** get_value ***************/
static char *nat_get_shape = "\
_set_thread_return_code (_self,ft_thread_get_value (%s,%s,%s));";

char *Get_Value(item_list_t list)
{
   char *event, *num, *result;
   item_cell_t cell;
   if (_length_of_item_list (list) != 3){
      return "error: bas number of arguments to get_value\n";
   }
   cell = _first_cell (list);
   event = _content_of_cell (cell);
   num = _content_of_cell (_next_cell (cell));
   result = _content_of_cell (_next_cell (_next_cell (cell)));

   if (native) return mysprintf (nat_get_shape,event,num,result);
   return mysprintf ("STATE_GET_VALUE (%d,%s,%s,%s) {}",new_state (),event,num,result);
}

/*************** unlink ***************/
char *Unlink(void)
{
   if (native) return mysprintf ("_set_thread_return_code (_self,ft_thread_unlink ());");
   return "error: non-native module cannot unlink\n";
}

/*************** link ***************/
static char *aut_link_shape = "\
STATE (%d) {\n\
   if (_get_thread_scheduler (_self) != %s){\n\
      _register_unlink_order (_self);\n\
      _set_thread_scheduler (_self,%s);\n\
   }\n\
   GOTO_NEXT;\n\
}\n\
STATE (%d) {}";

static char *nat_link_shape = "\
{if (ft_thread_is_linked (_self)) ft_thread_unlink (); _set_thread_return_code (_self,ft_thread_link (%s));}";

char *Link(char *sched)
{
   if (native) return mysprintf (nat_link_shape,sched);
   return mysprintf (aut_link_shape,new_state(),sched,sched,new_state ());
}

/*************** if ***************/
static char *aut_if_shape = "\
STATE (%d) {if(!(%s)){IMMEDIATE (%d);}}\n\
%s\nSTATE (%d) {IMMEDIATE(%d);}\n\
%s\nSTATE (%d) {}";

char *If(int state,char *cond,char *th,int elstate,char *el)
{
   int end;
   if (native) return mysprintf ("if(%s){%s}else{%s}",cond,th,el);
   end = new_state ();
   return mysprintf (aut_if_shape,state,cond,elstate+1,th,elstate,end,el,end);
}

/*************** while ***************/
static char *aut_while_shape = "\
STATE (%d) {if (!(%s)) IMMEDIATE (%d);}\n\
%s\n\
STATE (%d) {IMMEDIATE (%d);}\n\
STATE (%d) {}";

char *While(int state,char *exp,char *body)
{
   int redo,end;
   if (native) return mysprintf ("while(%s){%s}",exp,body);
   redo = new_state ();
   end  = new_state();
   return mysprintf (aut_while_shape,state,exp,end,body,redo,state,end);
}

/*************** repeat ***************/
int InitRepeat (void)
{
   int res = new_state ();
   new_state (); // reserve state number
   return res;
}

static char *aut_repeat_shape = "\
STATE (%d) {local(%s) = %s;}\n\
STATE (%d) {if (local(%s)<=0) IMMEDIATE(%d);}\n\
%s\n\
STATE (%d) {local(%s)--; IMMEDIATE (%d); }\n\
STATE (%d) {}";

char *Repeat_aut(int st,char *num,char *body)
{
   char *count = next_local_var ();
   int decr    = new_state ();
   int end     = new_state ();

   add_local_data ("int",count);
   
   return mysprintf (aut_repeat_shape,
		     st,
		     count,num,
		     st+1,count,end,
		     body,
		     decr,count,st+1,
		     end);
}

static int rcount = 0;
static char *nat_repeat_shape = "\
{int _i%d = %s;\nfor(;_i%d>0;_i%d--){%s}\n}";

char *Repeat_native(int st,char *num,char *body)
{
   int count  = rcount++;
   return mysprintf (nat_repeat_shape,count,num,count,count,body);
}

char *Repeat(int st,char *num,char *body)
{
   if (native) return Repeat_native(st,num,body);
   return Repeat_aut(st,num,body);
}

/*************** run ***************/
static char *creation_shape = "\
   ft_thread_t th = %s_create_in (current_scheduler ()%s);\n\
   local(_run) = th;";

static char *creation (char *name,char *params)
{
   return mysprintf (creation_shape,
	     name,
	    (0 == strcmp (params,"")) ? "" : mysprintf (",%s",params));
}

static char *nat_run_shape = "\
{\n\
   %s\n\
   _set_thread_return_code (_self,ft_thread_join (th));\n\
   local(_run) = NULL;\n\
}";

static char *autom_run_shape = "\
STATE (%d) {\n\
   %s\n\
}\n\
STATE_JOIN (%d,local(_run)) {local(_run) = NULL;}";

char *Run(char *name,char *params)
{
   int first = new_state ();
   int second = new_state ();
   char *cr = creation(name,params);

   run_exists = 1;

   if (native) return mysprintf (nat_run_shape,cr);
   return mysprintf (autom_run_shape,first,cr,second);
}

