#include "rewrite_sem.h"

/********************************************/
#define SET_TYPE(x,v) x->parent.type = v

/********************************************/
static struct _instruction_t cooperate_ = {_COOPERATE};
static instruction_t cooperate_constant = &cooperate_;

instruction_t make_cooperate ()
{
   return cooperate_constant;
}

/********************************************/
static struct _instruction_t nothing_ = {_NOTHING};
static instruction_t nothing_constant = &nothing_;

instruction_t make_nothing ()
{
   return nothing_constant;
}

/********************************************/
static struct _instruction_t halt_ = {_HALT};
static instruction_t halt_constant = &halt_;

instruction_t make_halt ()
{
   return halt_constant;
}

/********************************************/
static struct _instruction_t return_ = {_RETURN};
static instruction_t return_constant = &return_;

instruction_t make_return ()
{
   return return_constant;
}

/********************************************/
instruction_t make_sequence (instruction_t left,instruction_t right)
{
   sequence_t res = mymalloc (sizeof (struct _sequence_t));
   SET_TYPE(res,_SEQ);
   res->left = left;
   res->right = right;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_stop (expression_t exp)
{
   stop_t res = mymalloc (sizeof (struct _stop_t));
   SET_TYPE(res,_STOP);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_suspend (expression_t exp)
{
   suspend_t res = mymalloc (sizeof (struct _suspend_t));
   SET_TYPE(res,_SUSPEND);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_resume (expression_t exp)
{
   resume_t res = mymalloc (sizeof (struct _resume_t));
   SET_TYPE(res,_RESUME);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_if (expression_t exp,instruction_t left,instruction_t right)
{
   if_t res = mymalloc (sizeof (struct _if_t));
   SET_TYPE(res,_IF);
   res->exp = exp;
   res->left = left;
   res->right = right;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_inst_from_exp (expression_t exp)
{
   inst_from_exp_t res = mymalloc (sizeof (struct _inst_from_exp_t));
   SET_TYPE(res,_EXP_INST);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_repeat (expression_t exp,instruction_t body)
{
   repeat_t res = mymalloc (sizeof (struct _repeat_t));
   SET_TYPE(res,_REPEAT);
   res->exp = exp;
   res->body = body;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_repeat_star (int n,instruction_t init,instruction_t body)
{
   repeat_star_t res = mymalloc (sizeof (struct _repeat_star_t));
   SET_TYPE(res,_REPEAT_STAR);
   res->counter = n;
   res->init = init;
   res->body = body;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_generate (expression_t exp)
{
   generate_t res = mymalloc (sizeof (struct _generate_t));
   SET_TYPE(res,_GENERATE);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_generate_valued (expression_t evt_exp,expression_t val_exp)
{
   generate_valued_t res = mymalloc (sizeof (struct _generate_valued_t));
   SET_TYPE(res,_GENERATE_VALUED);
   res->evt_exp = evt_exp;
   res->val_exp = val_exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_await (expression_t exp)
{
   await_t res = mymalloc (sizeof (struct _await_t));
   SET_TYPE(res,_AWAIT);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_await_star (event_t event)
{
   await_star_t res = mymalloc (sizeof (struct _await_star_t));
   SET_TYPE(res,_AWAIT_STAR);
   res->event = event;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_limited_await (expression_t evt_exp,expression_t limit_exp)
{
   limited_await_t res = mymalloc (sizeof (struct _limited_await_t));
   SET_TYPE(res,_LIMITED_AWAIT);
   res->evt_exp = evt_exp;
   res->limit_exp = limit_exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_limited_await_star (event_t event,int limit)
{
   limited_await_star_t res = mymalloc (sizeof (struct _limited_await_star_t));
   SET_TYPE(res,_LIMITED_AWAIT_STAR);
   res->event = event;
   res->limit = limit;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_atom (void *(*selector)(),int n)
{
   atom_t res = mymalloc (sizeof (struct _atom_t));
   SET_TYPE(res,_ATOM);
   res->selector = selector;
   res->n = n;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_join (expression_t exp)
{
   join_t res = mymalloc (sizeof (struct _join_t));
   SET_TYPE(res,_JOIN);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_limited_join (expression_t th_exp,expression_t limit_exp)
{
   limited_join_t res = mymalloc (sizeof (struct _limited_join_t));
   SET_TYPE(res,_LIMITED_JOIN);
   res->th_exp = th_exp;
   res->limit_exp = limit_exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_run (expression_t exp)
{
   run_t res = mymalloc (sizeof (struct _run_t));
   SET_TYPE(res,_RUN);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_while (expression_t exp,instruction_t body)
{
   while_t res = mymalloc (sizeof (struct _while_t));
   SET_TYPE(res,_WHILE);
   res->exp = exp;
   res->body = body;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_while_star (expression_t exp,instruction_t init,instruction_t body)
{
   while_star_t res = mymalloc (sizeof (struct _while_star_t));
   SET_TYPE(res,_WHILE_STAR);
   res->exp = exp;
   res->init = init;
   res->body = body;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_get_value (expression_t evt_exp,expression_t num_exp,void **var)
{
   get_value_t res = mymalloc (sizeof (struct _get_value_t));
   SET_TYPE(res,_GET_VALUE);
   res->evt_exp = evt_exp;
   res->num_exp = num_exp;
   res->var = var;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_get_value_star (event_t event,int k,void **var)
{
   get_value_star_t res = mymalloc (sizeof (struct _get_value_star_t));
   SET_TYPE(res,_GET_VALUE_STAR);
   res->event = event;
   res->num = k;
   res->var = var;
   return (instruction_t)res;
}
