#include <stdio.h>
#include "itemlist.h"
#include "atom.h"
#include "gramactions.h"

extern char *ModuleName;
char* mysprintf(char *format, ...);

int run_exists = 0;

int new_state (void) {return 0;}

/****************************************************/
char *numbefore (int n,char *txt)
{
   return mysprintf ("\n#line %d\n%s",n,txt);
}
char *numafter (char *txt,int n)
{
   return mysprintf ("%s\n#line %d\n",txt,n);
}

/****************************************************/
void InitAll (void)
{
   fprintf (stdout, "#include \"loft_sem.h\"\n");
   initialize_atom_list ();
}

/*************** sequence ***************/
char *Seq (char *left,char *right)
{
   return mysprintf("make_sequence (%s,%s)",left,right);
}

/*************** nothing ***************/
char *Nothing (void)
{
   return mysprintf ("make_nothing ()");
}

/*************** cooperate ***************/
char *Cooperate (void)
{
   return mysprintf ("make_cooperate ()");
}

/*************** halt ***************/
char *Halt (void)
{
   return mysprintf ("make_halt ()");
}

/*************** return ***************/
char *Return (void)
{
   return "make_return ()";
}
   
/*************** C Statements ***************/
char *Cstat(char *s)
{
   return mysprintf ("make_atom (%s_selector,%d)",ModuleName,add_atom (s));
}

/*************** call ***************/
char *Call (char *name,char *params)
{
   return Cstat (mysprintf ("%s(%s);",name,params));
}

/*************** finalizer ***************/
char *Final_Call (char *name,char *params)
{
   return mysprintf ("%s (%s);",name,params);
}

/*************** await ***************/
static char *await_shape = "\
make_await (make_wrapper (%s_selector,%d))";

static char *limited_await_shape = "\
make_limited_await (\
make_wrapper (%s_selector,%d),\
make_wrapper (%s_selector,%d))";

char *Await (item_list_t list)
{
   char *sig, *num;
   item_cell_t cell;
   int len = _length_of_item_list (list);
   
   if (len != 1 && len != 2){
      return "error: bas number of arguments to await\n";
   }
   cell = _first_cell (list);
   sig = _content_of_cell (cell);

   if (len == 1) return mysprintf  (await_shape,ModuleName,add_wrapper(sig));

   num = _content_of_cell (_next_cell (cell));
   return mysprintf (limited_await_shape,ModuleName,add_wrapper(sig),
		     ModuleName,add_wrapper(num));
}

/*************** join ***************/
static char *join_shape = "\
make_join (make_wrapper (%s_selector,%d))";

static char *limited_join_shape = "\
make_limited_join (\
make_wrapper (%s_selector,%d),\
make_wrapper (%s_selector,%d))";

char *Join (item_list_t list)
{
   char *th, *num;
   item_cell_t cell;
   int len = _length_of_item_list (list);   
   if (len != 1 && len != 2){
      return "error: bas number of arguments to join\n";
   }
   cell = _first_cell (list);
   th = _content_of_cell (cell);
   
   if (len == 1){
      return mysprintf (join_shape,ModuleName,add_wrapper(th));
   }
   num = _content_of_cell (_next_cell (cell));
   return mysprintf (limited_join_shape,ModuleName,
		     add_wrapper(th),ModuleName,add_wrapper(num));
}

/*************** get_value ***************/
static char *get_shape = "\
make_get_value (\
make_wrapper (%s_selector,%d),\
make_wrapper (%s_selector,%d),\
%s)";

char *Get_Value (item_list_t list)
{
   char *event, *num, *result;
   item_cell_t cell;
   if (_length_of_item_list (list) != 3){
      return "error: bas number of arguments to get_value\n";
   }
   cell = _first_cell (list);
   event = _content_of_cell (cell);
   num = _content_of_cell (_next_cell (cell));
   result = _content_of_cell (_next_cell (_next_cell (cell)));
   return mysprintf (get_shape,ModuleName,add_wrapper (event),
		     ModuleName,add_wrapper (num),result);
}

/*************** link, unlink ***************/
char *Unlink (void)
{
  return "error: link/unlink\n";
}

char *Link (char *sched)
{
  return "error: link/unlink\n";
}

/*************** if ***************/
static char *if_shape = "make_if (make_wrapper (%s_selector,%d),%s,%s)";

char *If (int state,char *cond,char *th,int elstate,char *el)
{
   return mysprintf (if_shape,ModuleName,add_wrapper(cond),th,el);
}

/*************** while ***************/
char *While (int state,char *exp,char *body)
{
   return mysprintf ("make_while (make_wrapper (%s_selector,%d),%s)",
		     ModuleName,add_wrapper (exp),body);
}

/*************** repeat ***************/
int InitRepeat (void) { return 0; }

char *Repeat (int state,char *num,char *body)
{
   return mysprintf ("make_repeat (make_wrapper (%s_selector,%d),%s)",
		     ModuleName,add_wrapper(num),body);
}

/*************** run ***************/
char *Run (char *name,char *params)
{
   run_exists = 1;
   return mysprintf ("make_run (make_wrapper (%s_selector,%d))",ModuleName,
		     add_wrapper (mysprintf ("%s_create (%s)",name,params)));
}

/***********************************/
