#include "status.h"
#include "loft_sem.h"

thread_t make_thread (scheduler_t sched,
		      instruction_t inst,
		      finalizer_t final,
		      void *locals)
{
   thread_t res      = mymalloc (sizeof (struct _thread_t));
   res->signal       = event_create_in (sched);
   res->run =
      make_sequence (
	 inst,
	 make_generate (make_const_expression (res->signal))
      );
   res->finalizer    = final;
   res->state        = _CONT;
   res->suspended    = 0;
   res->called       = NULL;
   res->code         = OK;
   res->locals       = locals;
   res->scheduler    = sched;
   return res;
}

/****************************************************/
int fireable (thread_t thread)
{
   return (!thread->suspended) && (_CONT == thread->state);
}

/****************************************************/
void return_processing (thread_t thread,char val)
{
   thread->state = (val == _RTRN) ? _TERM : val;
}

/****************************************************/
void finalize (thread_t thread)
{
   thread->state = _TERM;
   generate (thread->signal);
   thread->finalizer (thread->locals);
   if (thread->called) finalize (thread->called);
}

/****************************************************/
thread_t self (void)
{
   return current_scheduler()->current;
}

/****************************************************/
int return_code (void)
{
   return self()->code;
}

/****************************************************/
void reset_thread_state (thread_t thread)
{
   if ((!thread->suspended) && (_COOP == thread->state))
      thread->state = _CONT; 
}

/****************************************************/
void* local_data (thread_t thread)
{
   return thread->locals;
}
